package comunicacao.sender;

import comunicacao.data.segment.Segment;
import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.util.Hashtable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import comunicacao.data.SegmentsConstructor;
import comunicacao.data.WindowSender;
import comunicacao.data.segment.Header;
import comunicacao.data.segment.Segment;
import comunicacao.data.segment.SequenceNumber;
import comunicacao.data.segmentBuffer.SegmentBuffer;
import comunicacao.data.segmentBuffer.SegmentBufferFile;
import comunicacao.exception.HeaderDontFitException;
import comunicacao.packets.SendingThread;

public class SenderThread implements Runnable {
	private DatagramSocket socket;
	private byte[] ackReceiverBuffer;
	private DatagramPacket ackReceiverPacket;

	private SegmentsConstructor segmentBuffer;
	private WindowSender window;	
	private int segmentIndex;

	private int amountOfDataAcked;
	private int totalAmountOfData;

	// Upload rate calculations variables
	private double uploadRate;
	private int uploadedData;// Amount of data sent since last getUploadRate() call.
	private long lastTime;// The moment of time the last getUploadRate() was called.

	private ExecutorService threads;
	private Hashtable<SequenceNumber, SendingThread> threadMapping;

	private TimeoutAlgorith transmissionTimeout;


	public SenderThread(DatagramSocket socket, byte[] data, int length, InetAddress remoteIP, int remotePort, ExecutorService executor) {
		this.socket = socket;
		this.ackReceiverBuffer = new byte[1024];
		this.ackReceiverPacket = new DatagramPacket(this.ackReceiverBuffer, this.ackReceiverBuffer.length);

		try {
			this.segmentBuffer = new SegmentsConstructor(data, length, remoteIP, remotePort);
		} catch (HeaderDontFitException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		this.window = new WindowSender(0);
		this.segmentIndex = 0;

		this.amountOfDataAcked = 0;	
		this.totalAmountOfData = length;

		this.uploadRate = 0;
		this.uploadedData = 0;
		this.lastTime = System.currentTimeMillis();

		this.threads = executor;

		this.transmissionTimeout = new TimeoutAlgorith();
		
		this.threadMapping = new Hashtable<SequenceNumber, SendingThread>();
	}

	@Override
	public void run() {
		// TODO Auto-generated method stub
		Segment toBeSent;
		Header ackReceivedHeader;
		System.out.println("Sender initialized");

		while(!hasFinished()){
			toBeSent = null;
			ackReceivedHeader = null;

			do {
				// Verifying if there are segments to be sent
				System.out.println("Sender will veryfi if there are segments to be sent");
				toBeSent = this.nextSegment();
				if (toBeSent != null) {
					// If there are, send them
					// TODO: Create sending thread
					System.out.println("Sender found segment " + toBeSent.getSequenceNumber().value());
					SendingThread thread = new SendingThread(toBeSent,
							this.segmentIndex - 1, transmissionTimeout, socket,
							threadMapping);
					this.threadMapping.put(toBeSent.getSequenceNumber(), thread);
					this.threads.submit(thread);
				}

			} while (toBeSent != null);

			try {
				//Waiting for acks for the recent sent segments
				System.out.println("Sender is waiting for acks");
				this.socket.receive(this.ackReceiverPacket);

				//Extracting ack iformation
				ackReceivedHeader = new Header(this.ackReceiverPacket);
				if (ackReceivedHeader.getFlags().isAck()){
					//Acking the segment (and adjusting the window)
					this.ackSegment(ackReceivedHeader.getSeqNum());
				
				}

			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}


		}
		System.out.println("Sender is dieing");
	}

	/**
	 * Method for checking the upload rate of this connection
	 * @return
	 */
	public synchronized double getUdloadRate() {
		long elapsedTime = System.currentTimeMillis() - this.lastTime;
		this.lastTime = System.currentTimeMillis();
		this.uploadRate = (double) this.uploadedData / elapsedTime;
		this.uploadedData = 0;
		return this.uploadRate;
	}

	private synchronized void addUploadedData(int dataLength) {
		this.uploadedData += dataLength;
	}

	/**
	 * Method for acking a segment right after an ack with its sequence number is received
	 * @param acked, the segment to be acked
	 */
	private synchronized void ackSegment(SequenceNumber seqNum) {
		Segment acked = this.segmentBuffer.getSegment(seqNum);
		this.ackSegment(acked);
	}

	/**
	 * Method for acking a segment right after its ack is received
	 * @param acked, the segment to be acked
	 */
	private synchronized void ackSegment(Segment acked) {
		// If the ack is not a duplicate ack
		if (acked != null ) {
			synchronized (acked) {
				if (!acked.isAcked()){
					this.threadMapping.get(acked.getSequenceNumber()).recalculateTimeout();
					System.out.println("Ack for segment "
							+ acked.getSequenceNumber().value());
					acked.setAcked(true);
					// this.decrementFlightSize();
					int dataAmount = acked.getPacket().getLength() - Header.HEADER_SIZE;
					this.addUploadedData(dataAmount);
					this.amountOfDataAcked += dataAmount;
					System.out.println("Sender acked " + dataAmount + " bytes");

					this.window.ackReceived();
					this.slideWindow();					
				}
			}


		}
	}

	/**
	 * Method for getting the next segment that's in the window limits to be sent.
	 * @return the segment, if it is in the windowLimits, null otherwise
	 */
	private synchronized Segment nextSegment() {
		Segment returnMe;

		if (this.window.isInWindow(this.segmentIndex)) {// Is inside the window, all is right
			returnMe = this.segmentBuffer.getSegment(segmentIndex);
			this.segmentIndex++;
		} else if (this.segmentIndex < this.window.getWindowBase()){// Is behind the window, adjust it, then return the segment
			this.segmentIndex = this.window.getWindowBase();
			returnMe = this.segmentBuffer.getSegment(segmentIndex);
			this.segmentIndex++;
		} else {//Is after the window, nothing to be done
			returnMe = null;
		}



		return returnMe;
	}


	/**
	 * Method for sliding the window when a non duplicate ack is received
	 */
	private synchronized void slideWindow() {
		//TODO set that the window should stop moving if hasFinished()
		Segment aux = this.segmentBuffer.getSegment(this.window.getWindowBase());
		while (aux != null && aux.isAcked()) {			
			this.window.slideWindow();
			aux = this.segmentBuffer.getSegment(this.window.getWindowBase());
		}
	}

	/**
	 * Method for adjusting the sending rate when a Timeout occurs in one of the segments
	 */
	public synchronized void timeoutOccurred() {
		this.segmentIndex = this.window.getWindowBase() + 1;
		this.window.timeoutOccurred();
	}

	/**
	 * Tests if all the segments have been successfully acked
	 * @return
	 */
	public synchronized boolean hasFinished(){
		return this.window.getWindowBase() == this.segmentBuffer.getNumberOfSegments();
	}

	public synchronized int getTotalAmountOfDataAcked() {
		return amountOfDataAcked;
	}

	@SuppressWarnings("unused")
	private synchronized void setTotalAmountOfDataAcked(
			int totalAmountOfDataAcked) {
		this.amountOfDataAcked = totalAmountOfDataAcked;
	}

	public synchronized SegmentsConstructor getSegmentBuffer() {
		return segmentBuffer;
	}

	public synchronized void setSegmentBuffer(SegmentsConstructor segmentBuffer) {
		this.segmentBuffer = segmentBuffer;
	}

	public WindowSender getWindow(){
		return this.window;
	}


}
